# Author(s): Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Copyright 2022 Kristo Mela
# -*- coding: utf-8 -*-
"""
Created on Tue Dec 14 19:48:26 2021

Stressed skin design

@author: kmela
"""

import numpy as np

from metku.eurocodes.en1993.en1993_1_3.trapezoidal import TrapezoidalSheet
from metku.eurocodes.en1993.en1993_1_3 import en1993_1_3

from metku.sections.steel.RHS import RHS, SHS
from metku.structures.steel.steel_member import SteelMember

K1 = {0: [[0.013, 0.03, 0.041, 0.041, 0.046, 0.05, 0.066, 0.103, 0.193],
          [0.042, 0.096, 0.131, 0.142, 0.142, 0.153, 0.199, 0.311, 0.602],
          [0.086, 0.194, 0.264, 0.285, 0.283, 0.302, 0.388, 0.601, 1.188],
          [0.144, 0.323, 0.438, 0.473, 0.468, 0.494, 0.629, 0.972, 1.935],
          [0.216, 0.438, 0.654, 0.703, 0.695, 0.729, 0.922, 1.42, 2.837],
          [0.302, 0.674, 0.911, 0.98, 0.965, 1.008, 1.266, 1.938, 3.892],
          [0.402, 0.895, 1.208, 1.3, 1.277, 1.329, 1.661, 2.536, 5.098],
          [0.516, 1.146, 1.546, 1.662, 1.631, 1.692, 2.107, 3.208, 6.453]],
      5: [[0.014, 0.031, 0.041, 0.044, 0.044, 0.049, 0.066, 0.107, 0.205],
          [0.05, 0.099, 0.128, 0.134, 0.132, 0.146, 0.198, 0.336, 0.652],
          [0.107, 0.202, 0.253, 0.26, 0.254, 0.28, 0.386, 0.681, 1.548],
          [0.188, 0.338, 0.413, 0.417, 0.404, 0.448, 0.629, 1.158, 2.639],
          [0.295, 0.507, 0.604, 0.601, 0.578, 0.648, 0.934, 1.783], 
          [0.429, 0.706, 0.823, 0.806, 0.772, 0.877, 1.306, 2.586], 
          [0.591, 0.935, 1.066, 1.028, 0.983, 1.135, 1.756, 3.605], 
          [0.78, 1.191, 1.328, 1.264, 1.208, 1.423, 2.299, 4.838]],
      10: [[0.016, 0.031, 0.04, 0.042, 0.042, 0.048, 0.065, 0.111, 0.221],
           [0.056, 0.101, 0.123, 0.125, 0.123, 0.139, 0.2, 0.366, 0.873],
           [0.125, 0.204, 0.238, 0.233, 0.226, 0.264, 0.402, 0.786], 
           [0.222, 0.338, 0.375, 0.356, 0.345, 0.418, 0.689, 1.445], 
           [0.349, 0.494, 0.526, 0.486, 0.473, 0.605, 1.082, 2.428], 
           [0.502, 0.668, 0.682, 0.615, 0.608, 0.837, 1.607], 	
           [0.677, 0.851, 0.834, 0.736, 0.752, 1.128, 2.308], 	
           [0.869, 1.035, 0.975, 0.844, 0.907, 1.494, 3.2]],
      15: [[0.017, 0.031, 0.04, 0.041, 0.041, 0.047, 0.066, 0.115],
           [0.062, 0.102, 0.118, 0.115, 0.113, 0.134, 0.209, 0.403],
           [0.139, 0.202, 0.218, 0.204, 0.2, 0.254, 0.44, 0.945],
           [0.244, 0.321, 0.325, 0.293, 0.294, 0.414, 0.796], 
           [0.37, 0.448, 0.426, 0.371, 0.396, 0.636, 1.329], 
           [0.506, 0.568, 0.508, 0.434, 0.513, 0.941], 	
           [0.646, 0.668, 0.561, 0.483, 0.664, 1.349], 	
           [0.766, 0.735, 0.578, 0.527, 0.861]],
      20: [[0.018, 0.032, 0.039, 0.039, 0.039, 0.046, 0.066, 0.111],
           [0.068, 0.101, 0.111, 0.106, 0.104, 0.131, 0.221, 0.452],
           [0.148, 0.193, 0.194, 0.174, 0.177, 0.255, 0.492], 
           [0.249, 0.289, 0.267, 0.23, 0.259, 0.444, 0.931], 
           [0.356, 0.372, 0.315, 0.27, 0.364, 0.725], 	
           [0.448, 0.42, 0.326, 0.303, 0.512],
           [0.509, 0.423, 0.301, 0.346],
           [0.521, 0.372, 0.259, 0.413]],      
      25: [[0.019, 0.032, 0.038, 0.038, 0.038, 0.045, 0.068, 0.126],
           [0.072, 0.099, 0.103, 0.095, 0.095, 0.129, 0.236, 0.513],
           [0.151, 0.178, 0.166, 0.144, 0.16, 0.268, 0.557], 
           [0.238, 0.244, 0.204, 0.176, 0.247, 0.494],
           [0.306, 0.272, 0.203, 0.204, 0.376],
           [0.333, 0.248, 0.172, 0.241],
           [0.3, 0.174, 0.142],
           [0.204, 0.081]],
      30: [[0.02, 0.032, 0.037, 0.036, 0.036, 0.044, 0.07, 0.133],
           [0.075, 0.095, 0.094, 0.084, 0.087, 0.132, 0.256], 
           [0.148, 0.157, 0.135, 0.116, 0.152, 0.291], 	
           [0.208, 0.186, 0.139, 0.139, 0.253], 
           [0.226, 0.161, 0.112, 0.176], 
           [0.18, 0.089, 0.093],
           [0.077]],
      35: [[0.021, 0.032, 0.036, 0.034, 0.034, 0.043, 0.072, 0.142],
           [0.076, 0.089, 0.083, 0.072, 0.082, 0.137, 0.281], 
           [0.137, 0.13, 0.102, 0.093, 0.151], 
           [0.162, 0.119, 0.082, 0.12],
           [0.123, 0.059],
           [0.032]],
      40: [[0.023, 0.032, 0.034, 0.032, 0.032, 0.043, 0.075, 0.155],
           [0.075, 0.081, 0.07, 0.06, 0.077, 0.146], 	
           [0.116, 0.096, 0.068, 0.078], 	
           [0.1, 0.053, 0.048],
           [0.024]],
      45: [[0.024, 0.031, 0.032, 0.029, 0.03, 0.043, 0.079], 
           [0.071, 0.069, 0.056, 0.05, 0.073], 
           [0.086, 0.057, 0.041],  
           [0.032]]
      }

class StressedSkin:
    """ General class for stressed skins """
    
    
    def __init__(self,sheet,a,b,n,seam_fastener={'d':4.8, 'spacing':500, 'loc':'trough'}):
        """
        

        Parameters
        ----------
        sheet : TYPE
            DESCRIPTION.
        seam_fastener : TYPE
            DESCRIPTION.
        a : TYPE
            DESCRIPTION.
        b : TYPE
            DESCRIPTION.
        n : TYPE
            DESCRIPTION.

        Returns
        -------
        None.

        """
    
        self.sheet = sheet
        self.seam_fastener = seam_fastener
        
        self.a = a
        self.b = b
        self.n = n
                
        self.p = sheet.pitch
        
        self.fastener_calc = 'en1993'
        
        # These values are from Table 5.1 of ECCS (1995)
        # slip in the fastening between sheeting and rafter
        self.sp = 0.35
        # slip in the seam fasteners
        self.ss = 0.25
        # slip in edge members/shear connectors
        self.ssc = 0.35
    
    def Fs(self):
        """ Seam fastener strength """
        
        fu = self.sheet.fu
        d = self.seam_fastener['d']
        t = self.sheet.t
        
        if self.fastener_calc == 'en1993':        
            # Calculate resistance as bearing resistance according to
            # EN 1993-1-3
            t1 = self.sheet.t
            Fs = en1993_1_3.screw_bearing_resistance(fu,d,t,t1)[0]
        else:
            # Calculation according to ECCS guide
            Fs = min(2.9*np.sqrt(t/d)*fu*d*t,3.8e3)
            
        
        return Fs

    def beta3(self):
        """ Value of parameter beta3 which takes into account
            the location of seam fasteners
        """
        
        if self.seam_fastener['loc'] == 'trough':
            beta3 = 1.0
        else:
            beta3 = (self.nf-1)/self.nf

        return beta3

    def beta1(self):
        """ Factor to allow for the number of sheet/purlin
            fasteners per sheet width.
            
            ECCS (1995), Annex C.1
        """
        
        nf = self.nf()
        
        if (nf % 2) == 0:
            # Even number of fasteners
            if self.seam_fastener['loc'] == 'trough':
                # Decking
                beta1 = sum([((2*(i+1)-1)/(nf-1))**3 for i in range(0,int(nf/2))])
            else:
                # Sheeting
                beta1 = sum([((2*(i+1)-1)/nf)**3 for i in range(0,int(nf/2))])
        else:
            # Odd number of fasteners
            if self.seam_fastener['loc'] == 'trough':
                # Decking
                beta1 = sum([(2*(i+1)/(nf-1))**3 for i in range(0,int((nf-1)/2))])
            else:
                # Sheeting
                beta1 = sum([(2*(i+1)/nf)**3 for i in range(0,int((nf-1)/2))])
        
        return beta1

    def nf(self):
        """ Number of sheet/purlin fasteners per sheet width
            including those at the overlaps.
            
            This is equal to the number of corrugations
        """
        
        return self.sheet.ntrough
    
    def K(self,fasteners_in_every_trough=True,verb=False):
        """ Table 5.9 """
        
        d = self.sheet.pitch
        h = self.sheet.h
        l = self.sheet.bt
        
        l2d = l/d
        h2d = h/d
        
        lnd = int(l2d/0.1)-1
        hnd = int(h2d/0.1)-1
        
        angle = 90-self.sheet.angle
        
        int_angle = int(5*np.floor(angle/5))
        
        if fasteners_in_every_trough:
            K = K1[int_angle][hnd][lnd]
        else:
            raise ValueError("Fasteners in every other trough is not included yet.")
        
        if verb:
            print(f"l/d = {l2d:.2f}")
            print(f"h/d = {h2d:.2f}")
            print(f"angle = {angle:.2f}")
            print(f"rounded angle = {int_angle:.2f}")
            print(f"K = {K:.4f}")
        
        return K

class StressedSkinRafters(StressedSkin):
    """ Class for stressed skin design using rafters.
        The corrugations are spanning in the direction of the building.
    """
    
    def __init__(self,sheet,a,b,n,seam_fastener,edge_fastener,rafter,edge_member):
        """

        Parameters
        ----------
        sheet : TYPE
            DESCRIPTION.
        seam_fastener : TYPE
            DESCRIPTION.
        a : TYPE
            DESCRIPTION.
        b : TYPE
            DESCRIPTION.
        n : TYPE
            DESCRIPTION.
        rafter : TYPE
            DESCRIPTION.
        edge_member : TYPE
            DESCRIPTION.

        Returns
        -------
        None.

        """
        
        super().__init__(sheet,a,b,n,seam_fastener)
        
        self.rafter = rafter
        self.edge_member = edge_member
        self.edge_fastener = edge_fastener
        
    @property
    def ns(self):
        """ Number of seam fasteners per seam in one panel,
            excluding the seam-to-rafter fasteners.
        """
        return int(self.b/self.seam_fastener['spacing'])-1
    
    @property
    def nsc(self):
        """ Number of edge member fasteners in one panel,
            excluding the seam-to-rafter fasteners.
        """
        return int(self.b/self.edge_fastener['spacing'])
    
    @property
    def nsh(self):
        """ Number of sheet widths per shear panel """
        return round(self.a/self.sheet.width)
    
    @property
    def rafter_t(self):
        """ Return rafter wall thickness """
        
        if hasattr(self.rafter,'tf'):
            t = self.rafter.tf
        elif hasattr(self.rafter,'T'):
            t = self.rafter.T
        elif hasattr(self.rafter,'tt'):
            t = self.rafter.tt
        
        return t
    
    @property
    def edge_member_t(self):
        """ Return edge member wall thickness """
        
        if hasattr(self.edge_member,'tf'):
            t = self.edge_member.tf
        elif hasattr(self.edge_member,'T'):
            t = self.edge_member.T
        elif hasattr(self.edge_member,'tt'):
            t = self.edge_member.tt
        
        return t
    
    def Fp(self):
        """ Shear resistance of fastener between rafter and sheet """
        
        fu = self.sheet.fu
        d = self.seam_fastener['d']
        t = self.sheet.t
        
        if self.fastener_calc == 'en1993':        
            # Calculate resistance as bearing resistance according to
            # EN 1993-1-3            
            t1 = self.rafter_t
                        
            Fp = en1993_1_3.screw_bearing_resistance(fu,d,t,t1)[0]
        else:
            # Calculation according to ECCS guide
            if d == 5.5:
                Fpmax = 6.5e3
            elif d == 6.3:
                Fpmax = 8.0e3
            else:
                raise ValueError("ECCS Guide only allows 5.5 mm or 6.3 mm sheet-to-rafter fasteners.")
                
            Fp = min(1.9*fu*d*t,Fpmax)
        
        return Fp
    
    def Fsc(self):
        """ Shear resistance of fastener between edge member and sheet """
        
        fu = self.sheet.fu
        d = self.edge_fastener['d']
        t = self.sheet.t
        
        if self.fastener_calc == 'en1993':        
            # Calculate resistance as bearing resistance according to
            # EN 1993-1-3            
            t1 = self.edge_member_t
                        
            F = en1993_1_3.screw_bearing_resistance(fu,d,t,t1)[0]
        else:
            # Calculation according to ECCS guide
            if d == 5.5:
                Fmax = 6.5e3
            elif d == 6.3:
                Fmax = 8.0e3
            else:
                raise ValueError("ECCS Guide only allows 5.5 mm or 6.3 mm sheet-to-edge member fasteners.")
                
            F = min(1.9*fu*d*t,Fmax)
        
        return F
        
    def seam_strength(self,verb=False):
        """ Resistance of the seam """
        
        Fs = self.Fs()
        Fp = self.Fp()
        
        ns = self.ns
        beta1 = self.beta1()
        beta3 = self.beta3()
        
        Vs = self.a/self.b*(ns*Fs + beta1/beta3*Fp)
        
        if verb:
            print("Seam strength:")
            print(f"a = {self.a:.2f} mm")
            print(f"b = {self.b:.2f} mm")
            print(f"Fs = {Fs*1e-3:.2f} kN")
            print(f"Fp = {Fp*1e-3:.2f} kN")
            print(f"beta_1 = {beta1:.0f} ")
            print(f"beta_3 = {beta3:.0f} ")
            print(f"ns = {ns:.0f} ")
            print(f"Vs = {Vs*1e-3:.2f} kN")
                    
        return Vs

    def edge_member_fastener_strength(self,verb=False):
        """ ECCS (1995), 5.8.1.2 """
        
        Fsc = self.Fsc()
        nsc = self.nsc
        
        V = self.a/self.b*Fsc*nsc
        
        if verb:
            print("Edge member fastener strength:")
            print(f"Fsc = {Fsc*1e-3:.2f} kN")
            print(f"nsc = {nsc:.0f} ")
            print(f"V = {V*1e-3:.2f} kN")
        
    def sheet_rafter_fastener_strength(self,verb=False):
        """ ECCS (1995), 5.8.3.1 """
        
        V = 0.6*self.a*self.Fp()/self.p

        if verb:
            print("Sheet-to-rafter fastener strength:")
            print(f"a = {self.a:.2f} mm")
            print(f"p = {self.p:.2f} mm ")
            print(f"V = {V*1e-3:.2f} kN")
    
        return V

    def end_collapse(self,verb=False):
        """ ECCS (1995), 5.8.3.2 """
        
        t = self.sheet.t
        fy = self.sheet.fy
        a = self.a
        d = self.sheet.pitch
        
        every_corrugation_fastened = True
        
        if every_corrugation_fastened:
            K = 0.9
        else:
            K = 0.3
        
        V = K*t**1.5*a*fy/np.sqrt(d)
        
        if verb:
            print("End collapse resistance:")
            print(f"a = {a:.2f} mm")
            print(f"t = {t:.2f} mm ")
            print(f"fy = {fy:.2f} MPa ")
            print(f"d = {d:.2f} mm ")
            print(f"V = {V*1e-3:.2f} kN")

    def profile_distortion(self,verb=True):
        """ Shear flexibility for profile distortion,
            Table 5.9 of ECCS (1995)
        """
        
        a = self.a
        b = self.b
        t = self.sheet.t
        E = self.sheet.E*1e-3
        d = self.sheet.pitch
        
        K = self.K()
        
        # Factor to take into account sheet continuity
        alpha5 = 1.0
        
        c1_1 = a*d**2.5*alpha5*K/E/t**2.5/b**2
        
        if verb:
            print("Shear flexibility: profile distortion")
            print(f"a = {a:.2f} mm")
            print(f"b = {b:.2f} mm")
            print(f"t = {t:.2f} mm ")            
            print(f"d = {d:.2f} mm ")
            print(f"K = {K:.4f} ")
            print(f"alpha_5 = {alpha5:.1f} ")
            print(f"c1_1 = {c1_1:.4f} mm/kN")
        
        return c1_1

    def shear_strain(self,verb=True):
        """ Shear flexibility for shear strain,
            Table 5.9 of ECCS (1995)
        """
        
        a = self.a
        b = self.b
        t = self.sheet.t
        h = self.sheet.h
        E = self.sheet.E*1e-3
        nu = self.sheet.nu
        d = self.sheet.pitch
                        
        c1_2 = 2*a*(1+nu)*(1+2*h/d)/E/t/b
        
        if verb:
            print("Shear flexibility: shear strain")            
            print(f"c1_2 = {c1_2:.4f} mm/kN")

        return c1_2

    def fastener_deformation(self,verb=True):
        """ Fastener deformation components """
               
        sp = self.sp
        ss = self.ss
        ssc = self.ssc
        nsh = self.nsh
        ns = self.ns
        nsc = self.nsc
        beta1 = self.beta1()
        
        # Sheet-to-rafter
        c2_1 = 2*self.a*sp*self.p/self.b**2
        
        # Seam fasteners
        c2_2 = ss*sp*(nsh-1)/(ns*sp+beta1*ss)
        
        # Edge members
        c2_3 = 2*ssc/nsc
        
        if verb:
            print("Shear flexibility: sheet-to-rafter")            
            print(f"c2_1 = {c2_1:.4f} mm/kN")
            print("Shear flexibility: seam fasteners")            
            print(f"c2_2 = {c2_2:.4f} mm/kN")
            print("Shear flexibility: edge members")  
            print(f"c2_3 = {c2_3:.4f} mm/kN")
        
        return [c2_1,c2_2,c2_3]
        
    def edge_member_strain(self,verb=True):
        """ Axial strain in edge members """
        a = self.a
        b = self.b
        E = self.edge_member.E*1e-3
        A = self.edge_member.A
        n = self.n
                
        c3 = n**2*b**3/E/A/a**2
        
        if verb:
            print("Shear flexibility: axial strain of edge member")            
            print(f"c3 = {c3:.4f} mm/kN")
        
        return c3
    
    def shear_flexibility(self,verb):
        """ Shear flexibility of one panel """
        
        c1_1 = self.profile_distortion(verb)
        c1_2 = self.shear_strain(verb)
        c2 = self.fastener_deformation(verb)
        c3 = self.edge_member_strain(verb)
        
        c_prime = self.b**2/self.a**2*(c1_1+c1_2+sum(c2))
        
        c = c_prime + c3
        
        if verb:
            print("Shear flexibility: ")
            print(f"c_prime = {c_prime:.4f} mm/kN")
            print(f"c = {c:.4f} mm/kN")
        
        return c
        
        
        
if __name__ == "__main__":
    
    sheet = TrapezoidalSheet(1.0,130,111,75,61)
    rafter = RHS(160,160,6)
    edge_member = RHS(100,100,5)
    seam_fastener = {'d':4.8, 'spacing':500, 'loc':'trough'}
    edge_fastener = {'d':5.5, 'spacing':250}
    a = 24000
    b = 6000
    n = 12
    
    s = StressedSkinRafters(sheet, a, b, n, seam_fastener, edge_fastener, rafter, edge_member)    
    
    #s.seam_strength(True)
    #s.edge_member_fastener_strength(True)
    #s.sheet_rafter_fastener_strength(True)
    #s.end_collapse(True)
    
    #s.profile_distortion(True)
    #s.shear_strain(True)
    #s.fastener_deformation(True)
    
    s.shear_flexibility(True)
    


# -*- coding: utf-8 -*-
# Copyright 2022 Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Author(s): Kristo Mela
from metku.frame2d.frame2d import *
from metku.frame2d.geometry import *
from metku.frame2d.materials import MATERIALS
from metku.framefem.elements import *
from metku.framefem.framefem import *
from metku.sections.steel import *


class FrameMember(GeometryMember):
    """
    CLASS DEFINITION
    """
    num_members = 0

    def __new__(cls, p1, p2, **kwargs):
        """
        Checks whether the dimension of the coordinates is 2 or 3
        :param p1: start coordinate
        :param p2: end coordinate
        :param kwargs: optional keyword arguments
        :return: 2D or 3D FrameMember instance
        """
        dim = len(p1)
        if dim == 2:
            return FrameMember2D(p1, p2, **kwargs)
        elif dim == 3:
            return FrameMember3D(p1, p2, **kwargs)

    def __init__(self, start_coord, end_coord, **kwargs):
        super().__init__(start_coord, end_coord)

        # Default values
        values = dict(
            profile='IPE 100',
            num_elements=4,
            Sj1=np.inf,
            Sj2=np.inf,
            material='S355'
        )
        # Update values with given kwargs
        values.update(kwargs)

        # Assign unique id
        self.mem_id = self.num_members
        FrameMember.num_members += 1

        # Initialize nodes and elements
        self.nodes = {}
        self.elements = {}
        self.nels = values['num_elements']

        # Initialize cross-sectional properties
        self.cross_section = None
        self.steel_member = None  # TODO: CHANGE THIS NAME TO A GENERAL ONE
        self.material = MATERIALS[values['material']]
        self.profile = values['profile']

        # Initialize rotational stiffness
        self.Sj1 = values['Sj1']
        self.Sj2 = values['Sj2']

        # List of intersecting objects and nodes
        self.intersecting_objs = []
        self.intersecting_nodes = {}

        # Create nodes and elements
        self.generate_nodes()
        self.generate_elements()

    @property
    def num_elements(self):
        return self.nels

    @num_elements.setter
    def num_elements(self, val):
        """
        Change number of elements
        TODO!!
        """
        self.update()

    @property
    def nodal_coordinates(self):
        """
        Returns nodal coordinates as a list
        :return: nodal coordinate list
        """
        coords = [list(node.coord) for node in self.nodes.values()]
        intersect = [list(node.coord) for node in self.intersecting_nodes.values()]
        if len(intersect):
            for coord in intersect:
                coords.append(coord)
        return coords

    @property
    def fy(self):
        return self.material.fy

    @property
    def profile(self):
        return str(self.cross_section)

    @profile.setter
    def profile(self, val):
        """
        Changes member's profile to given value.
        Calculates new cross-sectional properties and sets these new values
        to member's elements.

        :param val: string, profile name e.g. 'IPE 100'
        """
        if isinstance(val, list) and len(val) == 5:
            h, b, tf, tw, r = val
            if isinstance(self.cross_section, CustomISection):
                self.cross_section.H = h
                self.cross_section.B = b
                self.cross_section.tf = tf
                self.cross_section.tw = tw
                self.cross_section.r = r
                self.cross_section.cross_section_properties()
            else:
                self.cross_section = CustomISection(h, b, tf, tw, r)
        else:
            val = val.upper()
            splitted_val = val.split(" ")
            profile_type = splitted_val[0]
            if profile_type == 'IPE' or profile_type == 'HE':
                try:
                    H = int(splitted_val[1])
                    catalogue = True
                except ValueError:
                    H = float(splitted_val[1])
                    catalogue = False

            elif profile_type == 'WI':
                vals = splitted_val[1].replace('X', '-').split('-')
                H = float(vals[0])
                b1 = float(vals[3])
                b2 = float(vals[5])
                tw = float(vals[1])
                tf1 = float(vals[2])
                tf2 = float(vals[4])

            else:
                vals = splitted_val[1].split('X')
                if len(vals) == 2:
                    H = float(vals[0])
                    T = float(vals[1])
                elif len(vals) == 3:
                    H = float(vals[0])
                    B = float(vals[1])
                    T = float(vals[2])
                else:
                    raise TypeError(f'{splitted_val[1]} is not valid profile')

            if profile_type == 'IPE':
                self.cross_section = IPE(H, self.fy)

            elif profile_type == 'WI':
                self.cross_section = WISection(H, tw, [b1, b2], [tf1, tf2],
                                               self.fy)

            elif profile_type == 'HE':
                if splitted_val[2] == 'A':
                    self.cross_section = HEA(H, self.fy)
                elif splitted_val[2] == 'AA':
                    self.cross_section = HEAA(H, self.fy)
                elif splitted_val[2] == 'B':
                    self.cross_section = HEB(H, self.fy)
                elif splitted_val[2] == 'C':
                    self.cross_section = HEC(H, self.fy)
                elif splitted_val[2] == 'M':
                    self.cross_section = HEM(H, self.fy)

            elif profile_type == 'CHS':
                self.cross_section = CHS(H, T, self.fy)

            elif profile_type == 'RHS':
                self.cross_section = RHS(H, B, T, self.fy)

            elif profile_type == 'SHS':
                self.cross_section = SHS(H, T, self.fy)
            else:
                raise ValueError(
                    '{} is not valid profile type!'.format(profile_type))

        # Change steel_member objects properties
        if self.steel_member:
            self.steel_member.profile = self.cross_section

    @GeometryMember.start_coord.setter
    def start_coord(self, val):
        self.c0 = val
        self.update()


    @GeometryMember.end_coord.setter
    def end_coord(self, val):
        self.c1 = val
        self.update()

    def update(self):
        """
        Updates nodal coordinates
        """
        self.update_coordinates()
        self.update_elements()

    @property
    def locs(self):
        """
        List of nodes' local coordinates
        """
        if len(self.nodes):
            locs = [self.p1.distance(node.coord) / self.length for node in self.nodes.values()]
            intersect_locs = [self.p1.distance(node.coord) / self.length for node in self.intersecting_nodes.values()]
            if len(intersect_locs):
                locs.extend(intersect_locs)
        else:
            locs = list(np.arange(0, 1, 1 / self.num_elements))
            locs.append(1)

        return sorted(set(locs))

    def generate_elements(self):
        """
        Generates elements
        """
        for i in range(len(self.locs) -1):
            n0 = self.nodes[i]
            n1 = self.nodes[i + 1]
            elem = EBSemiRigidBeam(n0, n1, self.cross_section, self.material)
            self.elements[i] = elem

    def update_elements(self):
        """
        Updates elements
        """
        elements = self.elements.values()
        nodes = self.nodes
        nodes.update(self.intersecting_nodes)
        nodes = list(nodes.values())
        nodes.sort(key=lambda n: self.p1.distance(n.coord))
        for i in range(len(self.locs) -1):
            try:
                elem = elements[i]
                elem.n1 = nodes[i]
                elem.n2 = nodes[i + 1]
            except:
                n0 = nodes[i]
                n1 = nodes[i + 1]
                elem = EBSemiRigidBeam(n0, n1, self.cross_section,
                                       self.material)
                self.elements[i] = elem


    def generate_nodes(self):
        """
        Generates nodes
        """
        for i, loc in enumerate(self.locs):
            x, y = self.to_global(loc)
            nid = self.mem_id * 100 + i
            node = FEMNode(nid, x, y)
            self.nodes[i] = node

    def plot_elements(self, show=True):
        """
        Plots elements and nodes
        """
        X = []
        Y = []
        for node in self.nodes.values():
            X.append(node.x)
            Y.append(node.y)
            plt.scatter(node.x, node.y)
            plt.text(node.x, node.y, node.nid)
        plt.plot(X, Y)
        if show:
            plt.show()


class FrameMember2D(FrameMember, GeometryMember2D):

    def __new__(cls, p1, p2, **kwargs):
        return GeometryMember2D.__new__(cls, p1, p2, **kwargs)

    def __init__(self, start_coord, end_coord, **kwargs):
        super().__init__(start_coord, end_coord, **kwargs)

    def update_coordinates(self):
        """
        Updates nodes' coordinates
        """
        all_nodes = self.nodes
        all_nodes.update(self.intersecting_nodes)
        all_nodes = list(all_nodes.values())
        all_nodes.sort(key=lambda n: self.p1.distance(n.coord))
        for i, loc in enumerate(self.locs):
            x, y = self.to_global(loc)
            node = all_nodes[i]
            node.x = x
            node.y = y



class FrameMember3D(FrameMember, GeometryMember3D):

    def __new__(cls, p1, p2, **kwargs):
        return GeometryMember3D.__new__(cls, p1, p2, **kwargs)

    def __init__(self, start_coord, end_coord, **kwargs):
        super().__init__(start_coord, end_coord, **kwargs)

    def generate_nodes(self):
        """
        Generates nodes
        """
        locs = list(np.arange(0, 1, 1 / self.num_elements))
        locs.append(1)

        for i, loc in enumerate(locs):
            x, y, z = self.to_global(loc)
            nid = self.mem_id * 100 + i
            node = FEMNode(nid, x, y, z)
            self.nodes[i] = node

    def update_coordinates(self):
        """
        Updates nodes' coordinates
        """
        for i, loc in enumerate(self.locs):
            x, y, z = self.to_global(loc)
            node = self.nodes[i]
            node.x = x
            node.y = y
            node.z = z

    def plot_elements(self, show=True):
        """
        Plots elements and nodes
        """
        X = []
        Y = []
        Z = []

        fig = plt.figure()
        ax = fig.gca(projection='3d')

        for elem in self.elements.values():
            n1, n2 = elem.nodes
            X.append(n1.x)
            Y.append(n1.y)
            Z.append(n1.z)
            ax.scatter(n1.x, n1.y, n1.z)
            ax.text(n1.x, n1.y, n1.z, n1.nid)
        X.append(n2.x)
        Y.append(n2.y)
        Z.append(n2.z)
        ax.scatter(n2.x, n2.y, n2.z)
        ax.text(n2.x, n2.y, n2.z, n2.nid)

        ax.plot(X, Y, Z)
        if show:
            plt.show()


if __name__ == '__main__':
    mem = FrameMember([0, 0], [0, 1000], num_elements=5, number_elements=3)
    print(mem.plot())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#  This source code is licensed under the MIT license. See LICENSE in the repository root directory.
#  Copyright (c) 2022. Metku team.
#  All rights reserved.

from enum import Enum, auto, StrEnum

class ObjectiveTypeEnum(StrEnum):
    MIN = auto()
    MAX = auto()

class ObjectiveEnum(Enum):
    WEIGHT = 1
    COST = 2



# Author(s): Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Copyright 2022 Kristo Mela
# -*- coding: utf-8 -*-
"""
Created on Wed Mar 14 22:42:48 2018

Class for general steel cross sections 
    NOTE: This could be a package with the following structure
        cross_section
            hollow_sections
                RHS
                SHS
                CHS
            I_sections


@author: kmela
"""

from abc import abstractclassmethod, ABCMeta

import numpy as np

from metku.eurocodes.en1993 import constants
from metku.eurocodes.en1993 import en1993_1_1
from metku.materials.steel_data import Steel

INFEASIBLE = 999


class SteelSection(metaclass=ABCMeta):
    """ Base class for cross-sections
    
    Specific cross sections are subclasses of this class
    
    """

    # Constructor
    def __init__(self,
                 material,
                 A,
                 I,
                 Au,
                 Wpl,
                 Wel,
                 Ashear,
                 Ned=0.0,
                 Med=0.0,
                 Ved=0.0,
                 Ted=0.0,
                 design_code=en1993_1_1):
        """

        Parameters
        ----------
        material : int or str
            Either yields strength or name of the material.
        A : float
            Cross-sectional area.
        I : List
            second moment of area with respect to principal axes.
            I[0] .. major axis
            I[1] .. minor axis
        Au : float
            circumferential area per unit length.
        Wpl : list
            plastic section modulus with respect to principal axes.
        Wel : list
            elastic section modulus with respect to principal axes..
        Ashear : float
            shear area.
        Ned : float, optional
            design value of axial force. The default is 0.0.
        Med : list or float, optional
            bending moment. The default is 0.0.
        Ved : list or float, optional
            shear force. The default is 0.0.
            Ved[0] .. shear force is along the major axis
            Ved[1] .. shear force is along the minor axis
        Ted: float, optional
            torque, The default is 0.0
        design_code : TYPE, optional
            DESCRIPTION. The default is en1993_1_1.

        Returns
        -------
        None.

        """

        # print(material)
        # if isinstance(material,str):
        #    self.material = Steel(material)
        # elif isinstance(material,(float,int)):
        #    self.material = Steel("S" + str(int(material)))

        self.material = material

        self.E = constants.E
        self.A = A
        self.I = np.asarray(I)
        self.Au = Au
        self.Wpl = np.asarray(Wpl)
        self.Wel = np.asarray(Wel)

        if isinstance(Ashear, list):
            self.Ashear = Ashear
        else:
            self.Ashear = [0.0, Ashear]

        self.Ned = Ned
        if not isinstance(Med, list):
            self.Med = [Med, 0.0]
        else:
            self.Med = Med

        if not isinstance(Ved, list):
            self.Ved = [0.0, Ved]
        else:
            self.Ved = Ved

        self.Ted = Ted

        self.imp_factor = None
        self.code = design_code
        self.density = constants.density
        self.rotate = False

        # Cost of section, €/kg (or $/kg, or...)
        self.unit_cost = 1.0

    def __repr__(self):
        return type(self).__name__

    def cost(self):
        """ Material cost of section, units: €/m """
        return self.weight() * self.unit_cost

    @property
    def material(self):
        return self.__material

    @material.setter
    def material(self, val):
        """ Sets the material of section """
        if isinstance(val, str):
            self.__material = Steel(val)
        elif isinstance(val, (float, int)):
            self.__material = Steel("S" + str(int(val)))

    @property
    def G(self):
        """ Shear modulus """
        return self.material.G

    @property
    def fy(self):
        """ Yield strength """

        return self.material.fy

    @property
    def fu(self):
        """ Ultimate strength """

        return self.material.fu

    @property
    def Iy(self):
        if self.rotate:
            return self.I[1]
        return self.I[0]

    @Iy.setter
    def Iy(self, val):
        if self.rotate:
            self.I[1] = val
        self.I[0] = val

    @property
    def Iz(self):
        if self.rotate:
            return self.I[0]
        return self.I[1]

    @Iz.setter
    def Iz(self, val):
        if self.rotate:
            self.I[0] = val
        self.I[1] = val

    @property
    def Wply(self):
        if self.rotate:
            return self.Wpl[1]
        return self.Wpl[0]

    @Wply.setter
    def Wply(self, val):
        if self.rotate:
            self.Wpl[1] = val
        self.Wpl[0] = val

    @property
    def Wplz(self):
        if self.rotate:
            return self.Wpl[0]
        return self.Wpl[1]

    @Wplz.setter
    def Wplz(self, val):
        if self.rotate:
            self.Wpl[0] = val
        self.Wpl[1] = val

    @property
    def Wely(self):
        if self.rotate:
            return self.Wel[1]
        return self.Wel[0]

    @Wely.setter
    def Wely(self, val):
        if self.rotate:
            self.Wel[1] = val
        self.Wel[0] = val

    @property
    def Welz(self):
        if self.rotate:
            return self.Wel[0]
        return self.Wel[1]

    @Welz.setter
    def Welz(self, val):
        if self.rotate:
            self.Wel[0] = val
        self.Wel[1] = val

    @property
    def section_factor(self):
        """ Section factor for unprotected profile,
            exposed to fire on all sides
        """
        return self.Au / self.A * 1e3

    @property
    def box_section_factor(self):
        """ Section factor when considering the
            profile as a box
        """
        return self.section_factor

    def shadow_effect(self):
        """ Shadow effect for fire, EN 1993-1-2 """

        return 1.0

    def iy(self):
        """ Radius of gyration of the major axis """
        return np.sqrt(self.Iy / self.A)

    def iz(self):
        """ Radius of gyration of the minor axis """
        return np.sqrt(self.Iz / self.A)

    def i0(self):
        """ Polar radius """
        return np.sqrt(self.iy() ** 2 + self.iz() ** 2)

    @property
    def eps(self):
        return self.code.epsilon(self.fy)

    @property
    def NRd(self):
        if self.Ned > 0:
            return self.code.compression_resistance(self.A, self.fy)

        else:
            return self.code.compression_resistance(self.A, self.fy)

    @property
    def VRd(self):
        if abs(self.Ashear[1]) < 1e-6 or self.fy < 1e-3:
            print("STEEL SECTION, ASHEAR: ", self.Ashear[1])
        return self.code.shear_resistance(self.Ashear[1], self.fy)

    @property
    def VyRd(self):
        if abs(self.Ashear[0]) < 1e-6 or self.fy < 1e-3:
            print("STEEL SECTION, ASHEAR: ", self.Ashear[0])
        return self.code.shear_resistance(self.Ashear[0], self.fy)

    @property
    def VzRd(self):
        if abs(self.Ashear[1]) < 1e-6 or self.fy < 1e-3:
            print("STEEL SECTION, ASHEAR: ", self.Ashear[1])
        return self.code.shear_resistance(self.Ashear[1], self.fy)

    @property
    def MRd(self):
        M, C = self.bending_resistance()
        return M

        """
        if self.C < 3:
            return self.code.bending_resistance(self.Wpl[0], self.fy)
        elif self.C == 3:
            return self.code.bending_resistance(self.Wel[0], self.fy)
        else:

            #  print("USING ELASTIC PROPERTIES FOR CROSS-SECTION CLASS 4")
            return self.code.bending_resistance(self.Wel[0], self.fy)

            #raise NotImplemented("Calculation of cross-section class 4 is not implemented yet")
        """

    @property
    def MyRd(self):
        M, C = self.bending_resistance(axis="y")
        return M

        """
        if self.C < 3:
            return self.code.bending_resistance(self.Wpl[0], self.fy)
        elif self.C == 3:
            return self.code.bending_resistance(self.Wel[0], self.fy)
        else:

            #  print("USING ELASTIC PROPERTIES FOR CROSS-SECTION CLASS 4")
            return self.code.bending_resistance(self.Wel[0], self.fy)
        """

    @property
    def MzRd(self):
        M, C = self.bending_resistance(axis="z")
        return M

        """
        if self.C < 3:
            return self.code.bending_resistance(self.Wpl[1], self.fy)
        elif self.C == 3:
            return self.code.bending_resistance(self.Wel[1], self.fy)
        else:

            #  print("USING ELASTIC PROPERTIES FOR CROSS-SECTION CLASS 4")
            return self.code.bending_resistance(self.Wel[1], self.fy)
        """

    @property
    def C(self):
        C = self.section_class()
        return C

    @property
    def epsilon(self):
        e = self.code.epsilon(self.fy)
        return e

    @abstractclassmethod
    def flange_class(self):
        """ Determines flange's class in compression

            Returns:
            ---------
            :return C: flange's cross-section class in compression
            :rtype C: int
        """

    @abstractclassmethod
    def web_class_comp(self):
        """ Determines web's class in compression

            Returns:
            ---------
            :return C: web's cross-section class
            :rtype C: int

        """

    @abstractclassmethod
    def web_class_bend(self):
        """ Setermines web's class in bending

        Returns:
        ---------
        :return C: web's cross-section class in bending
        :rtype C: int
        """

    @abstractclassmethod
    def web_class_comp_bend(self, Ned):
        """ Determines web's class in combined bending and compression

            Parameters:
            -----------
            :param Ned: compressive force [kN]
            :type Ned: float

            Returns:
            --------
            :return C: web's cross-section class in combined bending and compression
            :rtype C: int

        """

    def robot(self):
        # returns the name of the section that Robot Structural Analysis can identify.
        pass

    def abaqus(self):
        # writes cross-section data to a file that abaqus can read
        pass

    def weight(self):
        """ Weight per unit length kg/mm """
        w = self.A * constants.density
        return w

    def self_weight(self):
        """ self-weight of the section (N/mm) """
        return self.weight() * 9.81

    def stress_ratio(self):
        """ computes the stress ratio Psi required
        in cross-section classification of parts in
        compression and bending
        See. EN 1993-1-1, Table 5.2
        NOTE: compression is positive, thus
        the '-' sign in front of Ned.
        """

        stmax = -self.Ned / self.A - self.Med / self.Wel[0]
        stmin = -self.Ned / self.A + self.Med / self.Wel[1]

        p = stmin / stmax

        return p

    def sigma_com(self):
        """ Elastic compressive stress 
            This works for double symmetric cross-sections
        """
        sN = -self.Ned / self.A
        sMy = self.Med[0] / self.Wel[0]

        return sN + sMy

    def sigmaN(self):
        """ Axial stress from normal force """

        return self.Ned / self.A

    def sigmaM(self):
        """ Axial stress from bending moment """
        return self.Med[0] / self.Wel[0]

    def section_class(self, verb=False):
        """ Determines cross-section class """

        C = 1
        Med = self.Med[0]
        
        if self.Ned >= 0.0 and abs(Med) < 1e-4:
            # No bending and zero axial force or tension
            print("No axial force or tension and no bending")
            C = 1
        
        # Pure bending        
        elif abs(self.Ned) < 1e-4 and abs(Med) > 1e-4:
            if verb:
                print("Pure bending")
            Cflange = self.flange_class(verb)
            Cweb = self.web_class_bend(verb)
            C = max(Cweb, Cflange)

        # Compression
        elif self.Ned < 0.0 or Med == 0:
            # Pure compression
            if abs(Med) < 1e-4:
                if verb:
                    print("Pure compression")
                Cflange = self.flange_class(verb)
                Cweb = self.web_class_comp(verb)
                C = max(Cweb, Cflange)
            # Bending and compression
            else:
                if verb:
                    print("Bending and compression")
                # Flange is in compression
                Cflange = self.flange_class(verb)
                # Classify web as a part in compression and bending
                Cweb = self.web_class_comp_bend(self.Ned, verb)
                C = max(Cweb, Cflange)

        # Axial tension and bending
        elif self.Ned > 0.0 and abs(Med) > 1e-4:
            C = 1

        if verb:
            print("Section class = {0}".format(C))

        return C

    def axial_force_resistance(self, verb=False):
        if self.Ned >= 0:
            NRd = self.code.tension_resistance(self.A, self.fy)
        else:
            NRd = self.code.compression_resistance(self.A, self.fy)

        if verb:
            print("NRd = {0:4.2f} kN".format(NRd * 1e-3))

        return NRd

    def bending_resistance(self, C=0, axis="y", verb=False):
        # Bending resistance, Nmm
        if C == 0:
            C = self.section_class()

        if axis == "y":
            n = 0
        else:
            n = 1

        if C < 3:
            WRd = self.Wpl[n]
        elif C == 3:
            WRd = self.Wel[n]
        else:
            raise NotImplementedError(f"Cross-section class 4 not implemented in: {self}")
            # WRd = self.Wel[n]

        MRd = self.code.bending_resistance(WRd, self.fy)

        if verb:
            print("MRd = {0:4.2f} kNm".format(MRd * 1e-6))

        return MRd, C

    def elastic_bending_resistance(self, axis="y"):
        """ Elastic bending resistance """
        if axis == "y":
            n = 0
        else:
            n = 1
        return self.code.bending_resistance(self.Wel[n], self.fy)

    def plastic_bending_resistance(self, axis="y"):
        """ Plastic bending resistance """
        if axis == "y":
            n = 0
        else:
            n = 1
        return self.code.bending_resistance(self.Wpl[n], self.fy)

    def shear_force_resistance(self, C=0, axis="z", verb=False):
        if C == 0:
            C = self.section_class()

        if axis == "y":
            n = 0
        else:
            n = 1

        if C < 3:
            VRd = self.code.shear_resistance(self.Ashear[n], self.fy)
        elif C == 3:
            VRd = self.code.shear_resistance(self.Ashear[n], self.fy)
        else:
            VRd = self.code.shear_resistance(self.Ashear[n], self.fy)

        if verb:
            print("Shear area = {0:4.2f} mm2".format(self.Ashear[n]))
            print("VRd = {0:4.2f} kN".format(VRd * 1e-3))

        return VRd

    def moment_axial_force_interact(obj, UN, MRd):
        return None

        """
        def a = increase_M_and_N(self):
            "" Increases bending moment and axial force proportionally,
            until cross-section resistance is obtained. Used for classification
            
            Needs to be modified for Python!
            ""
            
            a0 = 1.0
            
            #%a = fzero(@MNInc,a0)
            aIter = []
            amax = 5
            while isempty(aIter)
                amax = 2*amax
                [f,aIter] = secant(@MNInc,[1.0,amax])
                   
            a = aIter(end)                        
        
        
        def res = MNInc(k):
            selfN.Med = k*self.Med
            selfN.Ned = k*self.Ned
                
            MNRd = selfN.MomentAxialForceInteract
            if MNRd > 0:
                res = selfN.Med/MNRd-1
            else
                res = 1
        """

    def section_resistance(self, axis='y', return_list=True, verb=False):
        """ Calculates resistance of cross-section
            Checks the following:
                Axial force
                Shear force
                Bending moment
                Interaction of axial force and bending moment
            output: r .. maximum of different utilization ratios
        """
        if axis == 'y':
            idx = 0
            MRd = self.MyRd
            Med = self.Med[0]
            Ved = self.Ved[1]
            VRd = self.VzRd
        else:
            idx = 1
            Med = self.Med[1]
            Ved = self.Ved[0]
            MRd = self.MzRd
            VRd = self.VyRd

        # Normal force
        UN = abs(self.Ned) / self.NRd
        # Shear force
        UV = abs(Ved) / VRd
        # Bending moment
        UM = abs(Med) / MRd

        if self.C < 3:
            MNRd = self.moment_axial_force_interact(UN, MRd)
            if MNRd > 0.0:
                UMN = abs(Med) / MNRd
            else:
                UMN = UN + UM
                # UMN = INFEASIBLE
        else:
            UMN = UN + UM

        if return_list:
            r = [UN, UV, UM, UMN]
        else:
            r = max([UN, UV, UM, UMN])

        if verb:
            print("Cross-section design: " + self.__repr__() + " S" + str(self.fy))
            print("NEd = {0:4.2f} kN; NRd = {1:4.2f} kN => UN = {2:4.2f}".format(self.Ned * 1e-3, self.NRd * 1e-3, UN))
            print("VEd = {0:4.2f} kN; VRd = {1:4.2f} kN => UV = {2:4.2f}".format(Ved * 1e-3, VRd * 1e-3, UV))
            print("MEd = {0:4.2f} kNm; MRd = {1:4.2f} kNm => UM = {2:4.2f}".format(Med * 1e-6, MRd * 1e-6, UM))

        return r



# Author(s): Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Copyright 2022 Kristo Mela
# -*- coding: utf-8 -*-

from metku.frame2d.frame2d import *
from metku.optimization.structopt import *


AISC_PROFILE_AREAS = [71.613, 90.968, 126.451, 161.29, 198.064, 252.258,
                      285.161, 363.225, 388.386, 494.193, 506.451, 641.289,
                      645.16, 792.256, 816.773, 939.998, 1008.385, 1045.159,
                      1161.288, 1283.868, 1374.191, 1535.481, 1690.319,
                      1696.771, 1858.061, 1890.319, 1993.544, 2019.351,
                      2180.641, 2283.705, 2290.318, 2341.931, 2477.414,
                      2496.769, 2503.221, 2696.769, 2722.575, 2896.768,
                      2961.284, 3096.768, 3206.445, 3303.219, 3703.218,
                      4658.055, 5141.925, 5503.215, 5999.988, 6999.986,
                      7419.34, 8709.66, 8967.724, 9161.272, 9999.98, 10322.56,
                      10903.204, 12129.008, 12838.684, 14193.52, 14774.164,
                      15806.42, 17096.74, 18064.48, 19354.8, 21612.86]


class FiftyTwoBarTruss(OptimizationProblem):
    # Problem parameters
    E = 207e3  # MPa
    rho = 7860e-9  # kg/mm^3
    Fx = 100e3  # N
    Fy = 200e3  # N
    L = 2000  # mm
    H = 3000  # mm

    properties = {
        'L': str(L) + " mm",
        'Fx': f'{Fx / 1e3 :.2f} kN',
        'Fy': f'{Fy / 1e3 :.2f} kN',
        'E': f'{E / 1e3 :.2f} MPa',
        'rho': f'{rho * 1e9 :.0f} kg/m3'
    }

    # Constraint limits
    sigma_max = 180  # MPa

    def __init__(self, prob_type='discrete'):
        super().__init__("FiftyTwoBarTruss")
        self.prob_type = prob_type
        self.structure = self.create_structure()
        self.create_variables()
        self.create_objective()
        self.create_constraints()

    def constraint_generator(self, mem):
        """
        Creates non-linear constraint functions fro each member
        :param mem: member for which the constraints are created
        :return: created functions
        """

        def compression_fun(x):
            return -mem.ned / (mem.A * mem.fy) - 1

        def tension_fun(x):
            return mem.ned / (mem.A * mem.fy) - 1

        # def disp_fun(x):
        #     displacements = mem.nodal_displacements.values()
        #     max_vals = [max(l[0:2]) for l in displacements]
        #     min_vals = [min(l[0:2]) for l in displacements]
        #     max_val = max(max_vals)
        #     min_val = min(min_vals)
        #     abs_max = max(max_val, abs(min_val))
        #     return abs_max / self.delta_max - 1

        return compression_fun, tension_fun  # , disp_fun

    def create_constraints(self):
        """
        Creates constraints
        :return:
        """
        # Initialize constraints as an empty list
        self.cons = []
        for var in self.vars:
            for mem in var.target["objects"]:
                if isinstance(mem, FrameMember):
                    compression_fun, tension_fun = self.constraint_generator(
                        mem)

                    comp_con = NonLinearConstraint(con_fun=compression_fun,
                                                   name="Compression "
                                                        + str(mem.mem_id),
                                                   parent=self)
                    comp_con.fea_required = True

                    tension_con = NonLinearConstraint(con_fun=tension_fun,
                                                      name="Tension "
                                                           + str(mem.mem_id),
                                                      parent=self)
                    tension_con.fea_required = True

                    # disp_con = NonLinearConstraint(con_fun=disp_fun,
                    #                                name='Displacement '
                    #                                + str(mem_id),
                    #                                parent=self)
                    # disp_con.fea_required = True
                    self.cons.append(comp_con)
                    self.cons.append(tension_con)
                    # self.cons.append(disp_con)

    def create_structure(self):
        """
        Creates the structure
        :return: structure
        """

        frame = Frame2D(num_elements=1)
        nodes = {}
        # Starting coordinates for nodes
        x = 0
        y = 0
        for i in range(20):
            if i % 4:
                x += self.L

            elif i == 0:
                x = 0
                y = 0
            else:
                x = 0
                y += self.H

            nodes[i] = [x, y]

        # Create bars
        self.vert_groups = []
        self.diag_groups = []
        self.horiz_groups = []
        for j in range(4):
            # Vertical bars
            vert_group = []
            for k in range(4):
                # Node indices
                idx1 = j * 4 + k
                idx2 = idx1 + 4
                n1 = nodes[idx1]
                n2 = nodes[idx2]
                beam = SteelBeam([n1, n2])
                frame.add(beam)
                vert_group.append(beam)
            # Diagonal bars
            diag_group = []
            for l in range(3):
                idx1 = j * 4 + l
                idx2 = idx1 + 1
                idx3 = idx1 + 4
                idx4 = idx1 + 5
                n1 = nodes[idx1]
                n2 = nodes[idx2]
                n3 = nodes[idx3]
                n4 = nodes[idx4]
                beam1 = SteelBeam([n2, n3])
                beam2 = SteelBeam([n1, n4])
                frame.add(beam1)
                frame.add(beam2)
                diag_group.append(beam1)
                diag_group.append(beam2)

            # Horizontal bars
            horiz_group = []
            for m in range(3):
                idx1 = j * 4 + m
                idx2 = idx1 + 1
                idx3 = idx1 + 4
                idx4 = idx1 + 5
                n3 = nodes[idx3]
                n4 = nodes[idx4]
                beam = SteelBeam([n3, n4])
                frame.add(beam)
                horiz_group.append(beam)

            self.vert_groups.append(vert_group)
            self.diag_groups.append(diag_group)
            self.horiz_groups.append(horiz_group)

        # Supports & Loads
        for i in range(4):
            # Support
            nsup = nodes[i]
            frame.add(XYHingedSupport(nsup))
            # Point load
            npload = nodes[i + 16]
            pl1 = PointLoad(npload, [self.Fx, 0, 0])
            pl2 = PointLoad(npload, [0, self.Fy, 0])
            frame.add(pl1)
            frame.add(pl2)

        # Change Young's modulus
        for mem in frame.members.values():
            mem.E = self.E
            mem.fy = self.sigma_max

        frame.hinge_joints()
        # frame.add_self_weight()
        frame.generate()
        return frame

    def create_variables(self):
        """
        Creates variables used in optimization

        Appends each variable to 'vars' 'list from where they can be accessed
        """
        self.vars = []
        groups = []
        for i in range(4):
            groups.append(self.vert_groups[i])
            groups.append(self.diag_groups[i])
            groups.append(self.horiz_groups[i])

        for i, group in enumerate(groups):
            name = 'A' + str(i + 1)
            if self.prob_type == "discrete":
                var = DiscreteVariable(name,
                                       profiles=AISC_PROFILE_AREAS,
                                       target={"property": "A",
                                               "objects": group})
            elif self.prob_type == "continuous":
                var = Variable(name,
                               lb=AISC_PROFILE_AREAS[0],
                               ub=AISC_PROFILE_AREAS[-1],
                               target={"property": "A",
                                       "objects": group})
            else:
                raise TypeError("Problem type must be either 'discrete' "
                                "or 'continuous")

            self.vars.append(var)

    def create_objective(self):
        """
        Creates the objective function
        """

        def objective(x):
            weight = 0
            for mem in self.structure.members.values():
                weight += self.rho * mem.A * mem.length

            return weight

        self.obj = objective


if __name__ == '__main__':
    from metku.optimization.solvers import *
    from metku.optimization.solvers.optsolver import DiscreteVNS
    from metku.optimization.result_exporter import ResultExporter
    import matplotlib.pyplot as plt

    problem = FiftyTwoBarTruss('discrete')
    solver = MISLP(move_limits=[0.1, 10], gamma=1e-2)
    #solver = DiscreteVNS(step_length=3)
    #x0 = [43, 18,  5, 42, 15,  5, 29, 16,  7, 19, 18,  9]
    x0 = [var.ub for var in problem.vars]
    solver.solve(problem, x0=x0, maxiter=5, log=True)
    problem(solver.X, prec=6)

    exporter = ResultExporter(problem, solver)
    exporter.to_csv()
    #to_csv(problem, solver)



# Author(s): Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Copyright 2022 Kristo Mela
# -*- coding: utf-8 -*-
"""
Created on Wed Mar 14 23:12:31 2018

Circular hollow sections

@author: kmela
"""

import math

from metku.eurocodes.en1993 import en1993_1_1
from metku.sections.steel.steel_section import SteelSection

class CHS(SteelSection):
    """ Circular hollow sections """

    def __init__(self,D,T,fy=355):
        """ Circular hollow section
            D -- diameter
            T -- wall thickness
            fy -- yield strength
        """
        self.D = D
        self.T = T
        
        A = self.area()
        Au = self.paint_area()
        I = self.second_moment()
        Ashear = self.shear_area()
        Wel = self.elastic_modulus()
        Wpl = self.plastic_modulus()
        self.It = self.torsion_modulus()
        self.Iw = 0.0

        super().__init__(fy,A,I,Au,Wpl,Wel,Ashear)
        

        self.imp_factor = [en1993_1_1.buckling_curve["c"],en1993_1_1.buckling_curve["c"]]

    def __repr__(self):
        return f"{type(self).__name__} {self.D:.0f}X{self.T:.1f}"

    def area(self):
        r = 0.5*self.D
        return math.pi*r**2 - math.pi*(r-self.T)**2

        
    def paint_area(self):
        return math.pi*self.D

    def shear_area(self):
        return 2*self.area()/math.pi


    def second_moment(self):
        r2 = 0.5*self.D
        r1 = r2-self.T
        return 0.25*math.pi*(r2**4-r1**4)

    def elastic_modulus(self):
        return self.second_moment()/(0.5*self.D)

    def plastic_modulus(self):
        D2 = self.D
        D1 = self.D-2*self.T
        return (D2**3-D1**3)/6

    
    def torsion_modulus(self):
        Iv2 = math.pi*self.D**4/32
        Iv1 = math.pi*(self.D-2*self.T)**4/32
        Iv = Iv2-Iv1
        return Iv

    def web_class_bend(self, verb=False):
        r = self.D/self.T
        cWeb = en1993_1_1.CHS_class(r,self.eps)
        return cWeb

    def web_class_comp(self, verb=False):
        r = self.D/self.T
        cWeb = en1993_1_1.CHS_class(r,self.eps)
        return cWeb
        
    def web_class_comp_bend(self, verb=False):
        r = self.D/self.T
        cWeb = en1993_1_1.CHS_class(r,self.eps)
        return cWeb

    def flange_class(self, verb=False):
        return 1
    
    def bending_resistance(self, C=0, verb=False):
        # Bending resistance, Nmm
        if C == 0:
            C = self.section_class()

        if C < 3:
            WRd = self.Wpl
        elif C == 3:
            WRd = self.Wel
        else:
            raise NotImplementedError(f"Cross-section class 4 not implemented in: {self}")
            # WRd = self.Wel[n]

        MRd = self.code.bending_resistance(WRd, self.fy)

        if verb:
            print("MRd = {0:4.2f} kNm".format(MRd * 1e-6))

        return MRd, C
    
    def MomentAxialForceInteract(self):
        """ This interaction formula is only for plastic design """
        MRd = self.BendingResistance(1);
        NRd = self.AxialForceResistance;
            
        """ EN 1993-1-1: 6.2.9.1(5)
            In class 1 or 2 use plastic design
        """
        n = abs(self.Ned)/NRd;

        if n > 1.0:
            MNRd = 0
        else:
            MNRd = MRd*min(1-n**1.7,1)

        return MNRd
    
    
def CHS_area(D,T):
    r = 0.5*D
    A = math.pi*r**2 - math.pi*(r-T)**2
    return A
        
def CHS_paint_area(D):
    Au = math.pi*D
    return Au

def CHS_shear_area(A):
    Ashear = 2*A/math.pi
    return Ashear

def CHS_second_moment(D,T):
    r2 = 0.5*D
    r1 = r2-T
    I = 0.25*math.pi*(r2**4-r1**4)
    return I

def CHS_elastic_modulus(D,I):
    Wel = I/(0.5*D)
    return Wel

def CHS_plastic_modulus(D,T):
    D2 = D
    D1 = D-2*T
    Wpl = (D2**3-D1**3)/6
    return Wpl

def CHS_torsion_modulus(D,T):
    Iv2 = math.pi*D**4/32
    Iv1 = math.pi*(D-2*T)**4/32
    Iv = Iv2-Iv1
    return Iv

class CustomCHS(CHS):
    """ Class for custom CHS profiles, mainly for optimization """
    
    def __init__(self,D,T,fy=355):
        
        super().__init__(D,T,fy)
    
    def __getattribute__(self, name):
        """ override the attribute access for those attributes
            that depend on the section dimensions and that are
            constant for rolled sections
        """
        if name == "A":
            if self.area() < 0:
                print(self.D, self.T)
                
            return self.area()
        elif name == "I":
            I = self.second_moment()
            #if I[1] < 0:
            #    print(self.h, self.b, self.tt, self.tb, self.tw)
            return self.second_moment()
        elif name == "Wel":
            return self.section_modulus()
        elif name == "Wpl":
            return self.plastic_section_modulus()
        elif name == "Ashear":
            return self.shear_area()
        elif name == "Au":
            return self.paint_area()
        elif name == "It":
            return self.torsion_modulus()
        else:
            return super().__getattribute__(name)
    
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#  This source code is licensed under the MIT license. See LICENSE in the repository root directory.
#  Copyright (c) 2022. Metku team.
#  All rights reserved.

# @Filename : steel_section
# @Date : 2022
# @Project: metku
# @AUTHOR : Jaakko Huusko

from dataclasses import dataclass

import metku.eurocodes.en1993.en1993_1_1 as EC3_1
from metku.design.cross_section import CrossSection


@dataclass
class SteelSection(CrossSection):

    @property
    def NcRd(self) -> float:
        """
        Compression strength of cross-section
        :return: float [N]
        """
        return EC3_1.compression_resistance(self.profile.A, self.material.fy)

    @property
    def NtRd(self) -> float:
        """
        Tensile strength of cross-section
        :return: float [N]
        """
        return EC3_1.tension_resistance(self.profile.A, self.material.fy)

    @property
    def VyRd(self) -> float:
        """
        Shear strength about y-axis
        :return:
        """
        return EC3_1.shear_resistance(self.profile.Avy, self.material.fy)

    @property
    def VzRd(self) -> float:
        """
        Shear strength about z-axis
        :return:
        """
        return EC3_1.shear_resistance(self.profile.Avz, self.material.fy)

    @property
    def MplyRd(self) -> float:
        """
        Plastinc bending moment strength about y-axis
        :return:
        """
        return EC3_1.bending_resistance(self.profile.Wply, self.material.fy)

    @property
    def MelyRd(self) -> float:
        """
        Elastic bending moment strength about y-axis
        :return:
        """
        return EC3_1.bending_resistance(self.profile.Wely, self.material.fy)

    @property
    def MplzRd(self) -> float:
        """
        Plastinc bending moment strength about y-axis
        :return:
        """
        return EC3_1.bending_resistance(self.profile.Wplz, self.material.fy)

    @property
    def MelzRd(self) -> float:
        """
        Elastic bending moment strength about y-axis
        :return:
        """
        return EC3_1.bending_resistance(self.profile.Welz, self.material.fy)
